/*
 * $Id: parser.c,v 1.6 2021/03/03 16:06:56 mark Exp $
 *
 *  NAME
 *      parser - provide parsing for vc control lines
 *
 *  SYNOPSIS
 *      int parse(char *buf, char **arg);
 *          buf - pointer to control line.  First char is ignored
 *          (assumed to hold control line indicator)
 *          arg - returned with character pointer to argument
 *          that parse is not prepared to handle
 *
 *      parse returns an action code, or error code, as appropriate
 *
 *      const char *geterrmsg(int errno);
 *          returns char pointer to text corresponding to error code
 *          passed in errno
 *
 *  DESCRIPTION
 *      parse handles control statements and calls evaluate for those
 *      keywords taking expressions as arguments.  For other keywords
 *      with arguments, a char pointer to the argument is returned for
 *      the client routine to handle.
 *
 *      evaluate implements a recursive descent parser for a grammer
 *      of the form:
 *
 *      boolexpr ::= <boolterm> [<orop> <boolterm>]*
 *      boolterm ::= <notfactor> [<andop> <notfactor>]*
 *      notfactor ::= ['!'] <relation>
 *      relation ::= <expr> [<relop> <expr>]
 *      expr ::= <term> {'+'|'-'} <expr> | <term>
 *      term ::= <factor> {'*'|'/'} <term> | <factor>
 *      factor ::= '(' <expr> ')' | <constant> | <variable> ['=' <expr>]
 *      constant ::= <number> | <string>
 *      number ::= <digit> [<digit>]*
 *      string ::= '"' [<character>]* '"'
 *      variable ::= <alpha> [<alphanumeric>]*
 *
 *  NOTES
 *      Depends on symtab for variable handling and lexer for lexical
 *      scanning.
 *
 *  MODIFICATION HISTORY
 *  Mnemonic        Rel Date    Who
 *  parse           1.0 020924  mpw
 *      Created.
 *
 * Copyright (C) 2002, 2003, 2004 Mark Willson.
 *
 * This file is part of the vc program.
 *
 * vc is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * vc is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with vc; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lexer.h"
#include "parser.h"
#include "symtab.h"

/* define keyword table */
static struct keyentry {
    char *name;
    int action;
} keytab[] = {
    { "def",A_NOP },
    { "set",A_NOP },
    { "if",0 },
    { "else",0 },
    { "elif",0 },
    { "end",0 },
    { "inc",A_INC },
    { "on",A_EXPON },
    { "off",A_EXPOFF },
    { "ctl",A_NOP },
    { "msg",A_MSG },
    { "err",A_ERR },
    { "dumpsym",A_DUMPSYM }
};

#define NKEY (sizeof(keytab)/sizeof(struct keyentry))

extern char cchar;              /* control character */
extern int debug;               /* set true if debug info required */

static int exec = TRUE;         /* set TRUE if control statements should
                                 * be executed */

static int elifcnt;             /* count of elif levels */

static int error = 0;           /* contains error code, if non-zero */

static char *errtab[] = {
    "no such message",
    "unable to define variable",
    "unterminated string",
    "illegal symbol",
    "variable expected",
    "operand types do not match",
    "illegal type in expression",
    "unable to assign value to variable",
    "string expected",
    "illegal expression",
    "attempt to divide by zero",
    "insufficient memory",
    "integer expected",
    "unknown keyword",
    "stack overflow",
    "stack underflow",
    "stack empty"
};

#define MAX_ENUM (sizeof(errtab)/sizeof(char *))

/* function prototypes for recursive functions */
VAR expr(void);
VAR boolexpr(void);

/*-------------------------------------------------------------------------
 * return message string matching error code
 */

const char *geterrmsg(int e)
{
    int i;

    i = e- E_NUMBER;
    if (i > MAX_ENUM || i < 0) i = 0;
    return errtab[i];
}


/*------------------------------------------------------------------------
 * stack related stuff
 */

#define STKSZ 10

static int stk[STKSZ];
static int sp = 0;

int push(i)
int i;
{
    if (sp < STKSZ)
        stk[sp++] = i;
    else
        error = E_STKOVR;
    return(0);
}

int pull()
{
    if (sp > 0 )
        return(stk[--sp]);
    else
        error = E_STKUDR;
    return(0);
}

int top()
{
    if (sp > 0)
        return(stk[sp-1]);
    else
        error = E_STKEMP;
    return(0);
}

/*------------------------------------------------------------------------
 * perform operations on variables
 */

void addvar(VAR *v,VAR*v1)
{
    int tlen;
    char *ns;

    if (v->type != v1->type) {
        error = E_TYPEDIFF; /* type mismatch */
    }
    else {
        switch (v->type) {
            case S_INT:
                v->val.ival += v1->val.ival;
                break;
            case S_STR:
                tlen = strlen(v->val.pval)+strlen(v1->val.pval);
                ns = (char *) malloc(tlen+1);
                strcpy(ns,v->val.pval);
                strcat(ns,v1->val.pval);
                free(v->val.pval);
                v->val.pval = ns;
                break;
            default:
                error = E_BADTYPE;
        }
    }
}

void subvar(VAR *v,VAR*v1)
{
    if (v->type != v1->type) {
        error = E_TYPEDIFF; /* type mismatch */
    }
    else {
        switch (v->type) {
            case S_INT:
                v->val.ival -= v1->val.ival;
                break;
            default:
                error = E_BADTYPE;
        }
    }
}

void mulvar(VAR *v,VAR*v1)
{
    if (v->type != v1->type) {
        error = E_TYPEDIFF; /* type mismatch */
    }
    else {
        switch (v->type) {
            case S_INT:
                v->val.ival *= v1->val.ival;
                break;
            default:
                error = E_BADTYPE;
        }
    }
}

void divvar(VAR *v,VAR*v1)
{
    if (v->type != v1->type) {
        error = E_TYPEDIFF; /* type mismatch */
    }
    else {
        switch (v->type) {
            case S_INT:
                if (v1->val.ival == 0) {
                    error = E_DIVZERO;
                    break;
                }
                v->val.ival /= v1->val.ival;
                break;
            default:
                error = E_BADTYPE;
        }
    }
}

void cmpvar(int op,VAR *v,VAR *v1)
{
    int result = 0, cmpval;

    if (v->type != v1->type) {
        error = E_TYPEDIFF;
        return;
    }
    if (v->type == S_STR) {
        cmpval = strcmp(v->val.pval,v1->val.pval);
        switch (op) {
            case '<':
                result = cmpval < 0;
                break;
            case '>':
                result = cmpval > 0;
                break;
            case S_EQ:
                result = cmpval == 0;
                break;
            case S_NEQ:
                result = cmpval != 0;
                break;
            case S_LTEQ:
                result = cmpval <= 0;
                break;
            case S_GTEQ:
                result = cmpval >= 0;
                break;
            default:
                error = E_BADEXP;
        }
    }
    else if (v->type == S_INT) {
        switch (op) {
            case '<':
                result = v->val.ival < v1->val.ival;
                break;
            case '>':
                result = v->val.ival > v1->val.ival;
                break;
            case S_EQ:
                result = v->val.ival == v1->val.ival;
                break;
            case S_NEQ:
                result = v->val.ival != v1->val.ival;
                break;
            case S_LTEQ:
                result = v->val.ival <= v1->val.ival;
                break;
            case S_GTEQ:
                result = v->val.ival >= v1->val.ival;
                break;
            default:
                error = E_BADEXP;
        }
    }
    else {
        error = E_BADTYPE;
    }

    v->type = S_INT;
    v->val.ival = result;
}

void andvar(VAR *v, VAR *v1)
{
    if (v->type != v1->type) {
        error = E_TYPEDIFF;
        return;
    }
    if (v->type == S_INT) {
        v->val.ival = (v->val.ival && v1->val.ival);
    }
    else {
        error = E_BADEXP;
    }
    return;
}

void orvar(VAR *v, VAR *v1)
{
    if (v->type != v1->type) {
        error = E_TYPEDIFF;
        return;
    }
    if (v->type == S_INT) {
        v->val.ival = (v->val.ival || v1->val.ival);
    }
    else {
        error = E_BADEXP;
    }
    return;
}

/*-------------------------------------------------------------------------
 * functions to implement the recursive descent parser
 */

VAR factor()
{
    VAR v;
    int t;
    int uminus = FALSE;
    char var[LEXTOKSIZ+1];

    v.type = S_UNDEF;

    switch (lexsym) {
        case '-':
            uminus = TRUE;
            lexmatch('-');
            if (lexsym == S_VAR) {
                getvar(lextok,&v);
                if (v.type != S_INT) {
                    error = E_BADEXP;
                }
                else {
                    v.val.ival = -v.val.ival;
                    lexer();
                    break;
                }
            }
            else if (lexsym != S_INT) {
                error = E_BADINT;
                break;
            }
            /* fall thru */
        case S_INT:
            t = strtol(lextok,NULL,0);
            if (uminus) t = -t;
            v.type = S_INT;
            v.val.ival = t;
            lexer();
            break;
        case S_VAR:
            if (getvar(lextok,&v)) {
                strcpy(var,lextok);
            }
            else {
                error = E_BADEF; /* unable to define/get variable */
                v.type = S_INT;
                v.val.ival = 0;
                lexer();
                break;
            }
            lexer();
            if (lexsym == '=') {
                lexmatch('=');
                v = boolexpr();
                if (v.type != S_UNDEF) setvar(var,&v);
            }
            break;
        case S_STR:
            v.type = S_STR;
            v.val.pval = strsave(lextok);
            if (v.val.pval == NULL) {
                error = E_NOMEM; /* no memory */
                v.type = S_INT;
                v.val.ival = 0;
            }
            lexer();
            break;
        case '(':
            lexmatch('(');
            v = boolexpr();
            lexmatch(')');
            break;
        case S_EOE:
        case S_EOF:
        case ';':
            break;
        case S_USTR:
            error = E_BADSTR;
            break;
        default:
            error = E_BADSYM;
    }
    return v;
}

VAR term()
{
    VAR v,v1;

    v1.type = S_UNDEF;
    v = factor();

    switch (lexsym) {
        case '*':
            lexmatch('*');
            v1 = term();
            mulvar(&v,&v1);
            break;
        case '/':
            lexmatch('/');
            v1 = term();
            divvar(&v,&v1);
            break;
    }
    /* free malloc'd string memory if necessary */
    if (v1.type == S_STR) free(v1.val.pval);
    return v;
}

VAR expr()
{
    VAR v,v1;

    v1.type = S_UNDEF;

    v = term();
    switch (lexsym) {
        case '+':
            lexmatch('+');
            v1 = expr();
            addvar(&v,&v1);
            break;
        case '-':
            lexmatch('-');
            v1 = expr();
            subvar(&v,&v1);
            break;
        case S_UNDEF:
            error = E_BADEXP;
            break;
    }
    /* free malloc'd string memory if necessary */
    if (v1.type == S_STR) free(v1.val.pval);
    return v;
}



VAR relation()
{
    VAR v,v1;
    int op;

    v1.type = S_UNDEF;

    v = expr();
    switch (lexsym) {
        case '<':
        case '>':
        case S_LTEQ:
        case S_GTEQ:
        case S_NEQ:
        case S_EQ:
            op = lexsym;
            lexer();            /* eat relop */
            v1 = expr();
            cmpvar(op,&v,&v1);
            break;
    }
    return v;
}

VAR notfactor()
{
    VAR v;
    int foundnot = FALSE;

    if (lexsym == '!') {
        lexmatch('!');
        foundnot = TRUE;
    }

    v = relation();
    if (foundnot) {
        if (v.type == S_INT) {
            v.val.ival = !v.val.ival;
        }
        else {
            error = E_BADEXP;
        }
    }
    return v;
}

VAR boolterm()
{
    VAR v,v1;

    v1.type = S_UNDEF;

    v = notfactor();
    switch (lexsym) {
        case '&':
            lexmatch('&');
            v1 = notfactor();
            andvar(&v,&v1);
            break;
    }
    if (v1.type == S_STR) free(v1.val.pval);
    return v;
}

VAR boolexpr()
{
    VAR v,v1;

    v1.type = S_UNDEF;

    v = boolterm();
    switch (lexsym) {
        case '|':
            lexmatch('|');
            v1 = boolterm();
            orvar(&v,&v1);
            break;
    }
    if (v1.type == S_STR) free(v1.val.pval);
    return v;
}

/* shared by parse() and evaluate() */
static VAR value;

/*-------------------------------------------------------------------------
 * evaluate expression
 */

int evaluate()
{
    lexer();

    while (lexsym != S_EOF && lexsym != S_EOE) {
        error = 0;
        value = boolexpr();
        if (error == 0) {
            if (lexsym == ';') {
                switch (value.type) {
                    case S_INT:
                        printf("%d\n",value.val.ival);
                        break;
                    case S_STR:
                        printf("%s\n",value.val.pval);
                        break;
                    default:
                        fprintf(stderr,"Error: illegal value type in result: %d\n",
                                value.type);
                }
            }
        }
        /* free string memory */
        if (value.type == S_STR) free(value.val.pval);
        lexer();
    }
    return error;

}

/*------------------------------------------------------------------------
 * peval - interface directly to evaluate for rdp testing
 */

int peval(char *s)
{
    init_lexer(s);
    return evaluate();
}

/*------------------------------------------------------------------------
 * parse a control line passed in buf
 * places a char pointer to anything interesting in p
 *
 * Returns action code, or error number (error number >= E_NUMBER)
 */

int parse(char *buf,char **p)
{
    char *s;
    int rval,i;

    s = buf+1;
    init_lexer(s);
    lexer();

    if (debug) fprintf(stderr,"keyword: %s, type %d s: %s\n",lextok,lexsym,s);
    for (i=0;i<NKEY;i++) if (strcmp(keytab[i].name,lextok) == 0) break;
    if (i >= NKEY) return(E_BADKEY);
    rval = keytab[i].action;
    switch (i) {
        case 0:   /* def statement (obsolete, vars defined as
                   * required) */
            if (exec) {
                lexer();
                while (lexsym == S_VAR) {
                    if (!defvar(lextok)) return(E_BADEF);
                    lexer();
                }
                if (lexsym != S_EOE) return(E_BADEF);
            }
            break;
        case 1:  /* set statement */
            if (exec) {
                if ((i=evaluate()) != 0) return(i);
            }
            break;
        case 2:  /* if statement */
            push(exec);
            if (exec) {
                if ((i=evaluate()) != 0) return(i);
                exec = value.val.ival;
            }
            break;
        case 3:  /* else statement */
            if (top()) exec = !exec;
            break;
        case 4:  /* elif statment */
            elifcnt++;
            if (top()) exec = !exec;
            push(exec);
            if (exec) {
                if ((i=evaluate()) != 0) return(i);
                exec = value.val.ival;
            }
            break;
        case 5:  /* end statement */
            do {
                exec = pull();
            } while (elifcnt--);
            elifcnt = 0;
            break;
        case 6:  /* inc statement */
            if (exec) {
                lexer();
                *p = lextok;
            }
            else
                rval = A_NOP;
            break;
        case 7:  /* on statement */
        case 8:  /* off statement */
            if (!exec) rval = A_NOP;
            break;
        case 9: /* ctl statement */
            if (exec) {
                lexer();
                cchar = lextok[0];
            }
            else
                rval = A_NOP;
            break;
        case 10: /* msg statement */
        case 11: /* err statement */
            if (!exec) rval = A_NOP;
            else {
                lexer();
                *p = lextok;
                if (lexsym != S_STR) return(E_STREXP);
            }
            break;
        case 12:
            break;
    }
    if (rval == 0)
        return((exec)?A_COPY:A_SKIP);
    else
        return(rval);
}
